@@ -14,6 +14,7 @@ import cv2
 import numpy as np
 from numpy import ndarray
 import backoff
+from more_itertools import grouper
 from typing import Tuple, List, Dict, Union, Any, Optional
 
 
@@ -25,19 +26,28 @@ async def __fetch_img_data(session: ClientSession, headers, tile_data, image_til
     end_y = start_y + tile_data['height']
     start_x = tile_data['x']
     end_x = start_x + tile_data['width']
-    proxy_auth = aiohttp.BasicAuth(username, password)
-    async with session.get(tile_data['url'], headers=headers, proxy=proxy, proxy_auth=proxy_auth) as response:
-        image_data = np.asarray(bytearray(await response.read()), dtype="uint8")
-        image_tiles[start_y:end_y, start_x:end_x] = cv2.imdecode(image_data, cv2.IMREAD_COLOR)
+    if proxy and username and password:
+        proxy_auth = aiohttp.BasicAuth(username, password)
+        async with session.get(tile_data['url'], headers=headers, proxy=proxy, proxy_auth=proxy_auth) as response:
+            image_data = np.asarray(bytearray(await response.read()), dtype="uint8")
+    else:
+        async with session.get(tile_data['url'], headers=headers) as response:
+            image_data = np.asarray(bytearray(await response.read()), dtype="uint8")
+    
+    image_tiles[start_y:end_y, start_x:end_x] = cv2.imdecode(image_data, cv2.IMREAD_COLOR)
 
 
 @backoff.on_exception(backoff.expo, aiohttp.web.HTTPServerError, max_time=60)
 async def __fetch_json(session: ClientSession, headers, url: str, proxy: str, username: str, password: str) -> dict:
     """Get JSON data from a URL."""
 
-    proxy_auth = aiohttp.BasicAuth(username, password)
-    async with session.get(url, headers=headers, proxy=proxy, proxy_auth=proxy_auth) as response:
-        return await response.json()
+    if proxy and username and password:
+        proxy_auth = aiohttp.BasicAuth(username, password)
+        async with session.get(url, headers=headers, proxy=proxy, proxy_auth=proxy_auth) as response:
+            return await response.json()
+    else:
+        async with session.get(url, headers=headers) as response:
+            return await response.json()
 
 
 async def __calculate_image_tiles(session: ClientSession, url: str, headers: dict, crop_x: int = None, crop_y: int = None,
@@ -98,19 +108,20 @@ async def __calculate_image_tiles(session: ClientSession, url: str, headers: dic
 async def __download_manager(url: str, headers: dict, x: int = None, y: int = None, width: int = None, height: int = None, proxy: str = None, username: str = None, password: str = None, connector=None) -> ndarray:
     """Asynchronously download the image at the URL in the highest possible resolution."""
 
-    async with aiohttp.ClientSession(connector) as session:
+    async with aiohttp.ClientSession(connector=connector) as session:
         # Collect the tile data and initialize the necessary lists
         tiles, _, _, _, corr_width, corr_height = await __calculate_image_tiles(session, url, headers, x, y, width, height, proxy, username, password)
 
         # Fetch the tiles async
         image_shape = (corr_height, corr_width, 3)
         image_tiles = np.empty(image_shape, dtype=np.uint8)
-        tasks = []
-        for image_details in tiles:
-            tasks.append(__fetch_img_data(session, headers, image_details, image_tiles, proxy, username, password))
+        for tile_group in grouper(tiles, 4):
+            tasks = []
+            for image_details in [x for x in tile_group if x is not None]:
+                tasks.append(__fetch_img_data(session, headers, image_details, image_tiles, proxy, username, password))
 
-        # Wait for all requests to complete successfully
-        await asyncio.gather(*tasks)
+            # Wait for all requests to complete successfully
+            await asyncio.gather(*tasks)
         return image_tiles
 
 
